# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from . import Dataset
from climetlab.utils.domains import domain_to_area
from climetlab import load_source


class Era5SingleLevels(Dataset):
    def __init__(self, variable, period, domain=None, time=None, grid=None):

        request = dict(variable=variable, product_type="reanalysis",)

        if domain is not None:
            request["area"] = domain_to_area(domain)

        if time is not None:
            request["time"] = time

        if grid is not None:
            if isinstance(grid, (int, float)):
                request["grid"] = [grid, grid]
            else:
                request["grid"] = grid

        if isinstance(period, int):
            period = (period, period)

        sources = []
        for year in range(period[0], period[1] + 1):
            request["year"] = year
            sources.append(
                load_source("cds", "reanalysis-era5-single-levels", **request)
            )

        self.source = load_source("multi", sources)


dataset = Era5SingleLevels



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

"""
ipython is not None when running a notebook
"""

active = None
try:
    from IPython import get_ipython

    active = get_ipython()
except Exception:
    pass


def _identity(x, **kwargs):
    return x


if active:
    from IPython.display import display, Image, SVG, HTML, Markdown
else:
    Image = _identity
    SVG = _identity
    HTML = _identity
    Markdown = _identity
    display = _identity

__all__ = ["display", "Image", "SVG", "HTML", "Markdown"]



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from collections import defaultdict
import datetime
from dateutil.parser import parse

# datetime.fromisoformat() only available from Python3.7
from backports.datetime_fromisoformat import MonkeyPatch

MonkeyPatch.patch_fromisoformat()


def parse_date(date):
    return datetime.datetime.fromisoformat(date)


def to_datetime(dt):
    if isinstance(dt, datetime.datetime):
        return dt

    if isinstance(dt, str):
        return parse(dt)

    raise Exception("Unsupported date/time object %s (%s)" % (dt, type(dt)))


def to_datetimes_list(datetimes):
    if isinstance(datetimes, (datetime.datetime, str)):
        return to_datetimes_list([datetimes])

    if isinstance(datetimes, (list, tuple)):
        return [to_datetime(x) for x in datetimes]

    return datetimes


def _date_as_request(date):
    return "%d-%02d-%02d" % (date.year, date.month, date.day)


def _time_as_request(time):
    assert time.second == 0
    return "%02d:%02d" % (time.hour, time.minute)


def _indentity(x):
    return x


def datetimes_to_dates_and_times(datetimes_list, as_request=False):

    result = []
    datetimes = defaultdict(set)

    if as_request:
        _d = _date_as_request
        _t = _time_as_request
    else:
        _d = _indentity
        _t = _indentity

    for dt in to_datetimes_list(datetimes_list):
        datetimes[_d(dt.date())].add(_t(dt.time()))

    timedates = defaultdict(set)
    for date, times in sorted(datetimes.items()):
        times = tuple(sorted(times))
        timedates[times].add(date)

    for times, dates in timedates.items():
        result.append((tuple(sorted(dates)), times))

    return sorted(result)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


class PandasHelper:
    def __init__(self, frame, margins=0, **kwargs):

        self.frame = frame
        self.margins = margins

        if "lat@hdr" in self.frame:
            self.lat = "lat@hdr"
            self.lon = "lon@hdr"
        elif "lat" in self.frame:
            self.lat = "lat"
            self.lon = "lon"
        else:
            self.lat = "latitude"
            self.lon = "longitude"

    def plot_map(self, driver):

        margins = driver.option("margins", self.margins)
        column = driver.option("column", self.lat)

        north = self.frame[self.lat].max()
        south = self.frame[self.lat].min()

        lons1 = self.frame[self.lon]
        east1 = lons1.max()
        west1 = lons1.min()

        lons2 = self.frame[self.lon] % 360
        east2 = lons2.max()
        west2 = lons2.min()

        if abs(east1 - west1) <= abs(east2 - west2):
            east, west = east1, west1
        else:
            east, west = east2, west2

        driver.bounding_box(
            north=north + margins,
            south=south - margins,
            west=west - margins,
            east=east + margins,
        )

        driver.plot_pandas(self.frame, self.lat, self.lon, column)

    def bounding_box(self):

        north, east = self.frame[[self.lat, self.lon]].max()
        south, west = self.frame[[self.lat, self.lon]].min()

        return [
            north + self.margins,
            west - self.margins,
            south - self.margins,
            east + self.margins,
        ]

    def dates(self):
        return sorted(set([str(x).split("T")[0] for x in self.frame["time"].values]))


helper = PandasHelper



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from . import Dataset
from climetlab import load_source


class WeatherBench(Dataset):
    """
    This is an attempt to reproduce this research: https://arxiv.org/abs/2002.00469.
    See https://raspstephan.github.io/blog/weatherbench/
    There is a notebook available at: https://binder.pangeo.io/v2/gh/pangeo-data/WeatherBench/master?filepath=quickstart.ipynb
    """

    home_page = "https://raspstephan.github.io/blog/weatherbench/"
    citation = """@article{rasp2020weatherbench,
  title={WeatherBench: A benchmark dataset for data-driven weather forecasting},
  author={Rasp, Stephan and Dueben, Peter D and Scher, Sebastian and Weyn, Jonathan A and Mouatadid, Soukayna and Thuerey, Nils},
  journal={arXiv preprint arXiv:2002.00469},
  year={2020}
}
"""

    def __init__(self, parameter="geopotential_500", resolution=5.625):
        URL = "https://dataserv.ub.tum.de/s/m1524895/download?path=%2F{resolution}deg%2F{parameter}&files={parameter}_{resolution}deg.zip".format(
            resolution=resolution, parameter=parameter
        )
        self.source = load_source("url", URL, unpack=True)


dataset = WeatherBench



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from importlib import import_module
import os
import weakref
from climetlab.utils.html import table
from climetlab.core.caching import cache_file


def lookup(name):
    source = import_module(".%s" % (name.replace("-", "_"),), package=__name__)
    return source.source


def load(name, *args, **kwargs):
    source = lookup(name)(*args, **kwargs)
    if source.name is None:
        source.name = name
    return source


def list_entries():
    here = os.path.realpath(os.path.dirname(__file__))
    result = []

    for n in os.listdir(here):
        if n.startswith("."):
            continue

        if n.startswith("_"):
            continue

        if not n.endswith(".py"):
            continue

        result.append(n[:-3])

    return result


class DataSource:
    """
    Doc
    """

    name = None
    home_page = "-"
    licence = "-"
    documentation = "-"

    _dataset = None

    def __init__(self, **kwargs):
        self._kwargs = kwargs

    def cache_file(self, *args, extension=".cache"):
        owner = self.name
        if self.dataset:
            owner = self.dataset.name
        if owner is None:
            owner = self.__class__.__name__.lower()
        return cache_file(owner, *args, extension=extension)

    @property
    def dataset(self):
        if self._dataset is None:
            return None
        return self._dataset()

    @dataset.setter
    def dataset(self, dataset):
        self._dataset = weakref.ref(dataset)

    def _repr_html_(self):
        return table(self)

    def read_csv_options(self):
        if self.dataset is None:
            return {}
        return self.dataset.read_csv_options()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import entrypoints
import climetlab
from importlib import import_module
import logging

LOG = logging.getLogger(__name__)

CACHE = {}

PLUGINS = {}


def _load_plugins(kind):
    plugins = {}
    for e in entrypoints.get_group_all(f"climetlab.{kind}s"):
        plugins[e.name.replace("_", "-")] = e
    return plugins


def load_plugins(kind):
    if PLUGINS.get(kind) is None:
        PLUGINS[kind] = _load_plugins(kind)
    return PLUGINS[kind]


def find_plugin(directory, name, loader):

    plugins = load_plugins(loader.kind)

    if name in plugins:
        plugin = plugins[name]
        return loader.load_entry(plugin)

    n = len(directory)
    for path, _, files in os.walk(directory):
        path = path[n:]
        for f in files:
            if f.endswith(".yaml"):
                if f[:-5] == name:
                    full = os.path.join(directory, path, f)
                    return loader.load_yaml(full)

            if f.endswith(".py"):
                p = os.path.join(path, f[:-3])
                if p[0] != "/":
                    p = "/" + p
                if p[1:].replace("/", "-").replace("_", "-") == name:
                    return loader.load_module(p.replace("/", "."))

    raise Exception("Cannot find %s '%s'" % (loader.kind, name))


def directories():
    result = [os.path.dirname(climetlab.__file__)]
    for kind in ("dataset", "source"):
        for _, v in load_plugins(kind).items():
            try:
                module = import_module(v.module_name)
                result.append(os.path.dirname(module.__file__))
            except Exception:
                LOG.error("Cannot load module %s", v.module_name, exc_info=True)
    return result



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from . import Reader


class CSVReader(Reader):
    def to_pandas(self, **kwargs):
        import pandas

        options = dict()
        options.update(self.source.read_csv_options())
        options.update(kwargs)

        return pandas.read_csv(self.path, **options)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
from . import Reader

LOG = logging.getLogger(__name__)

try:
    from climetlab.grib_bindings import GribFile

    class GReader(GribFile):
        pass

    LOG.info("Using eccodes C bindings to decode GRIB data")

except AttributeError:  # eccodes not installed
    import pyeccodes

    class GReader(pyeccodes.Reader):
        def at_offset(self, offset):
            self.seek(offset)
            return next(self)

    LOG.info("Using pyeccodes to decode GRIB data")


class GribField:
    def __init__(self, handle, path):
        self.handle = handle
        self.path = path

    @property
    def values(self):
        return self.handle.get("values")

    def plot_map(self, driver):
        driver.bounding_box(
            north=self.handle.get("latitudeOfFirstGridPointInDegrees"),
            south=self.handle.get("latitudeOfLastGridPointInDegrees"),
            west=self.handle.get("longitudeOfFirstGridPointInDegrees"),
            east=self.handle.get("longitudeOfLastGridPointInDegrees"),
        )
        driver.plot_grib(self.path, self.handle.get("offset"))

    def to_numpy(self):
        return self.values.reshape((self.handle.get("Nj"), self.handle.get("Ni")))

    @property
    def offset(self):
        return int(self.handle.get("offset"))

    def __repr__(self):
        return "GribField(%s,%s,%s,%s,%s,%s)" % (
            self.handle.get("shortName"),
            self.handle.get("levelist"),
            self.handle.get("date"),
            self.handle.get("time"),
            self.handle.get("step"),
            self.handle.get("number"),
        )

    def grid_definition(self):
        return dict(
            north=self.handle.get("latitudeOfFirstGridPointInDegrees"),
            south=self.handle.get("latitudeOfLastGridPointInDegrees"),
            west=self.handle.get("longitudeOfFirstGridPointInDegrees"),
            east=self.handle.get("longitudeOfLastGridPointInDegrees"),
            south_north_increment=self.handle.get("jDirectionIncrementInDegrees"),
            west_east_increment=self.handle.get("iDirectionIncrementInDegrees"),
        )

    def metadata(self):
        m = dict()
        for n in ("shortName", "units"):
            p = self.handle.get(n)
            if p is not None:
                m[n] = str(p)

        return m


class GRIBIterator:
    def __init__(self, path):
        self.path = path
        self.reader = GReader(path)

    def __repr__(self):
        return "GRIBIterator(%s)" % (self.path,)

    def __next__(self):
        return GribField(next(self.reader), self.path)


class GRIBReader(Reader):
    def __init__(self, source, path):
        super().__init__(source, path)
        self._fields = None
        self._reader = None

    def __repr__(self):
        return "GRIBReader(%s)" % (self.path,)

    def __iter__(self):
        return GRIBIterator(self.path)

    def _items(self):
        if self._fields is None:
            self._fields = []
            for f in self:
                self._fields.append(f.offset)
        return self._fields

    def __getitem__(self, n):
        if self._reader is None:
            self._reader = GReader(self.path)
        return GribField(self._reader.at_offset(self._items()[n]), self.path)

    def __len__(self):
        return len(self._items())

    def to_xarray(self):
        import xarray as xr

        return xr.open_dataset(self.path, engine="cfgrib")



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import climetlab


def __getattr__(name):
    def wrapped(*args, **kwargs):
        return climetlab.load_dataset(name, *args, **kwargs)

    return wrapped



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from climetlab.core.plugins import directories
import os
import yaml
from collections import defaultdict
import logging
from climetlab.utils.html import css

LOG = logging.getLogger(__name__)


YAML_FILES = None


def _guess(data):
    if "dataset" in data:
        return "datasets"

    if "magics" in data:

        if "msymb" in data["magics"]:
            return "styles"

        if "mcount" in data["magics"]:
            return "styles"

        if "mcoast" in data["magics"]:
            return "layers"

        if "mmap" in data["magics"]:
            return "projections"

    return "unknown"


class Entry:
    def __init__(self, name, kind, path, data):
        self.name = name
        self.kind = kind
        self.path = path
        self.data = data
        self.hidden = data.get("hidden", False)

    def _repr_html_(self):
        html = [css("table")]
        html.append("<table class='climetlab'>")
        html.append("<tr><td>Name:</td><td>%s</td></tr>" % self.name)
        html.append("<tr><td>Collection:</td><td>%s</td></tr>" % self.kind)
        html.append("<tr><td>Path:</td><td>%s</td></tr>" % self.path)
        html.append("<tr><td>Definition:</td><td><pre>%s</pre></td></tr>" % (yaml.dump(self.data, default_flow_style=False),))
        html.append("</table>")
        return "".join(html)



def _load_yaml_files():
    global YAML_FILES
    if YAML_FILES is not None:
        return YAML_FILES

    YAML_FILES = defaultdict(dict)
    for root, _, files in os.walk(*directories()):
        for file in [f for f in files if f.endswith(".yaml")]:
            path = os.path.join(root, file)
            try:
                with open(path) as f:
                    data = yaml.load(f.read(), Loader=yaml.SafeLoader)
                    name, _ = os.path.splitext(os.path.basename(path))
                    kind = _guess(data)
                    collection = YAML_FILES[kind]
                    if name in collection:
                        LOG.warning(
                            "Duplicate entry for %s %s (using %s, ignoring %s)",
                            kind,
                            name,
                            collection[name].path,
                            path,
                        )
                    else:
                        collection[name] = Entry(name, kind, path, data)

            except Exception:
                LOG.error("Cannot process YAML file %s", path, exc_info=True)

    return YAML_FILES


def get_data_entry(kind, name):
    return _load_yaml_files()[kind][name]


def data_entries(kind=None):
    if kind is None:
        for collection in _load_yaml_files().items():
            for entry in collection.values():
                if not entry.hidden:
                    yield entry
    else:
        for entry in _load_yaml_files().get(kind, {}).values():
            if not entry.hidden:
                yield entry



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from . import Meteonet
from climetlab import load_source


class MeteonetMasks(Meteonet):
    """
    See https://github.com/meteofrance/meteonet
    """

    def __init__(
        self, domain="NW",
    ):

        url = "{url}/masks/{domain}_masks.grib".format(url=self.URL, domain=domain,)
        self.source = load_source("url", url, styles=["land-sea-mask", "orography"])


dataset = MeteonetMasks



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import sys
from importlib import import_module

HELPERS = {
    "xarray.core.dataset.Dataset": "xarray",
    "xarray.core.dataarray.DataArray": "xarray",
    "numpy.ndarray": "ndarray",
    "pandas.core.frame.DataFrame": "pandas",
    "builtins.NoneType": "none",
}


def helper(data, *args, **kwargs):
    """
    Returns an object that wraps classes from other packages
    to support
    """

    fullname = ".".join([data.__class__.__module__, data.__class__.__qualname__])

    name = HELPERS.get(fullname)

    if name is not None:
        helper = import_module(".%s" % (name,), package=__name__)
        try:
            return helper.helper(data, *args, **kwargs)
        except Exception:
            print(helper, file=sys.stderr)
            raise

    raise ValueError("Cannot find a helper for class %s" % (fullname,))
